@@ -23,10 +23,6 @@ has current_frame => (
     is => 'rw',
 );
 
-has pad => (
-    is => 'rw',
-);
-
 has lines => (
     is => 'rw',
 );
@@ -35,10 +31,6 @@ has code_len => (
     is => 'rw',
 );
 
-has macro => (
-    is => 'rw',
-);
-
 has function => (
     is => 'rw',
 );
@@ -47,8 +39,12 @@ has local_stack => (
     is => 'rw',
 );
 
+sub pad {
+    my($st) = @_;
+    return $st->frame->[ $st->current_frame ];
+}
 
-sub pc_arg {
+sub op_arg {
     $_[0]->{ code }->[ $_[0]->{ pc } ]->{ arg };
 }
 
